/*
  ######################################################################
  # THIS IS THE STRIPPED VERSION OF THE ORIGINAL FILE                  #
  ######################################################################
  # This file was taken from libTiMidity 0.1.0 tar-ball distribution   #
  # Read the file "README.TXT" in this directory for more details      #
  ######################################################################
  # September 14, 2005 - Aloysius Indrayanto                           #
  # All comments (except the library copyright & license) are stripped #
  ######################################################################
  # See the VortexGE documentation for information about VortexGE      #
  ######################################################################
*/

/*
    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <toivonen[AT]clinet[DOT]fi>
    Copyright (C) 2004 Konstantin Korikov <lostclus[AT]ua[DOT]fm>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef TIMIDITY_H
#define TIMIDITY_H

#include <stdio.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C"
{
#endif

#define LIBTIMIDITY_VERSION_MAJOR 0L
#define LIBTIMIDITY_VERSION_MINOR 1L
#define LIBTIMIDITY_PATCHLEVEL    0L

#define LIBTIMIDITY_VERSION \
        ((LIBTIMIDITY_VERSION_MAJOR<<16)| \
         (LIBTIMIDITY_VERSION_MINOR<< 8)| \
         (LIBTIMIDITY_PATCHLEVEL))

#define MID_AUDIO_U8        0x0008  
#define MID_AUDIO_S8        0x8008  
#define MID_AUDIO_U16LSB    0x0010  
#define MID_AUDIO_S16LSB    0x8010  
#define MID_AUDIO_U16MSB    0x1010  
#define MID_AUDIO_S16MSB    0x9010  
#define MID_AUDIO_U16       MID_AUDIO_U16LSB
#define MID_AUDIO_S16       MID_AUDIO_S16LSB

  typedef unsigned char uint8;
  typedef signed char sint8;
  typedef unsigned short uint16;
  typedef signed short sint16;
  typedef unsigned int uint32;
  typedef signed int sint32;

  typedef size_t (*MidIStreamReadFunc) (void *ctx, void *ptr, size_t size,
          size_t nmemb);
  typedef int (*MidIStreamCloseFunc) (void *ctx);

  typedef struct _MidIStream MidIStream;
  typedef struct _MidDLSPatches MidDLSPatches;
  typedef struct _MidSong MidSong;

  typedef struct _MidSongOptions MidSongOptions;
  struct _MidSongOptions
  {
    sint32 rate;    
    uint16 format;    
    uint8 channels;    
    uint16 buffer_size;    
  };

  typedef enum
  {
    MID_SONG_TEXT = 0,
    MID_SONG_COPYRIGHT = 1
  } MidSongMetaId;

  extern int mid_init (char *config_file);

  extern int mid_init_no_config (void);

  extern void mid_exit (void);

  extern MidIStream *mid_istream_open_file (const char *file);

  extern MidIStream *mid_istream_open_fp (FILE * fp, int autoclose);

  extern MidIStream *mid_istream_open_mem (void *mem, size_t size,
             int autofree);

  extern MidIStream *mid_istream_open_callbacks (MidIStreamReadFunc read,
             MidIStreamCloseFunc close,
             void *context);
  extern size_t mid_istream_read (MidIStream * stream, void *ptr, size_t size,
          size_t nmemb);
  extern void mid_istream_skip (MidIStream * stream, size_t len);

  extern int mid_istream_close (MidIStream * stream);

  extern MidDLSPatches *mid_dlspatches_load (MidIStream * stream);

  extern void mid_dlspatches_free (MidDLSPatches * patches);

  extern MidSong *mid_song_load (MidIStream * stream,
         MidSongOptions * options);

  extern MidSong *mid_song_load_dls (MidIStream * stream,
             MidDLSPatches * patches,
             MidSongOptions * options);

  extern void mid_song_set_volume (MidSong * song, int volume);

  extern void mid_song_start (MidSong * song);

  extern size_t mid_song_read_wave (MidSong * song, void *ptr, size_t size);

  extern void mid_song_seek (MidSong * song, uint32 ms);

  extern uint32 mid_song_get_total_time (MidSong * song);

  extern uint32 mid_song_get_time (MidSong * song);

  extern char *mid_song_get_meta (MidSong * song, MidSongMetaId what);

  extern void mid_song_free (MidSong * song);

#ifdef __cplusplus
}
#endif
#endif

