// ############################################################################################
//  guess-ext.c
//  Used by VortexGE configuration script to check if MMX, SSE, SSE2, SSE3 or 3DNow!
//  is supported by the current CPU
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

#include <stdio.h>
#include <stdint.h>


// Get the CPU capabilities
// This code was originally taken (with some modifications) from :
//     bsconf.c
//     uSTL library, ustl-0.8.tar.bz2
//     http://sourceforge.net/projects/ustl
#if defined(__GNUC__) && (__i386__ || __x86_64__)
inline bool cpuid_supported()
{
  unsigned long forig, fnew;

  // Pop flags, toggle ID bit, push, pop. cpuid supported if changed.
  asm
  (
    "pushf             \n\t"
    "pop        %0     \n\t"
    "mov        %0, %1 \n\t"
    "xor $0x200000, %0 \n\t"
    "push       %0     \n\t"
    "popf              \n\t"
    "pushf             \n\t"
    "pop        %0         "
    : "=r"(fnew), "=r"(forig)
  );
  return(fnew != forig);
}
#define i_cpuid(a,r,c,d) asm ("cpuid" : "=a"(r), "=c"(c), "=d"(d) : "0"(a) : "%ebx")
// Not GNUC and/or not an i386 nor x86_64, disable cpuid support.
#else
inline static bool cpuid_supported()
{
  return(false);
}
#define i_cpuid(a,r,c,d) r = 0; c = 0; d = 0
#endif


// Main program
int main()
{
  const uint32_t extFeatures   = 0x80000000;
  const uint32_t amdExtensions = 0x80000001;
  const uint32_t amdBits       = 0xC2400000;

  // Get the CPU capabilities
  if(!cpuid_supported()) return(0);

  uint32_t r, c, d;
  i_cpuid(0, r, c, d);
  if(!r) return(0);

  i_cpuid(1, r, c, d);              // Request for feature list
  uint32_t un32C1 = d & 0x17C99B31; // --- 0001 0111 1100 1001 1001 1011 0011 0001
  uint32_t un32C2 = c & 0x00000089; // --- 0000 0000 0000 0000 0000 0000 1000 1001
  uint32_t un32C3 = 0;

  i_cpuid(extFeatures, r, c, d);
  if(r != extFeatures) {
    i_cpuid(amdExtensions, r, c, d);
    un32C3 = d & amdBits;
  }

  // Check if all zeroes
  if((un32C1 | un32C2 | un32C3) == 0) return(0);

  // Extract and prints the CPU caps
  if(un32C1 & (uint32_t(1) << 23)) printf("_mmx_ ");
  if(un32C3 & (uint32_t(1) << 22)) printf("_emmx_ ");
  if(un32C1 & (uint32_t(1) << 25)) printf("_sse_ ");
  if(un32C1 & (uint32_t(1) << 26)) printf("_sse2_ ");
  if(un32C2 & (uint32_t(1) <<  0)) printf("_sse3_ ");
  if(un32C3 & (uint32_t(1) << 31)) printf("_3dnow_ ");
  if(un32C3 & (uint32_t(1) << 30)) printf("_e3dnow_ ");
  return(0);
}

