// ############################################################################################
//  Test.cc
//  A demo application for VortexGE
//  -------------------------------------------------------------------------------------------
//  NOTE: unlike other demo applications, this application is not created using
//        CBasicApplication class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

//#define __HIGH_RESOLUTION__

// Common headers for demo application
#include "DemoCommon.h"


// Initialization functions
bool InitializeSystem();
void LoadImages();
void LoadAudios();
void CreateScene();

// Rendering functions
void RenderPanels();
void RenderSprites();

// Objects
static CVideoEnv      csVidEnv;
static C3DEnv         cs3DEnv;
static CPanel3DBuffer csWorld;
static CMatrix3D      csMat;
static CImageBuilder  csImgBuild;
static CImage         csBackImg, csTreeImg;
static CAnimatedImage csLightImg, csArrowImg;
static CAImgControl   csArrowCtl;
static CTexture       csTex;
static CBitmapFont    csBF;
static CSprite2D      csTreeSpr, csLightSpr;
static CAudioEnv     &csAudioEnv =  CAudioEnv::AutoDetect();
static CSoundBuffer  &csSBBack   = *csAudioEnv.NewSoundBuffer(); //! Don't do this in real applications
static CSoundBuffer  &csSBShot   = *csAudioEnv.NewSoundBuffer(); //! Don't do this in real applications

// Rendering controls
static bool    bGouraudShaded, bBackImg, bTransparent, bFog;
static uint8_t un8TexType, un8ADir;


// Main program
int main()
{
  // Local variables
  CFPSControl csFPSCtrl(50.0f);  // For FPS control
  bool        bAlwaysOnTop;      // Always on top flag
  bool        bFullscreen;       // Full screen on top flag
  float       fALCol, fALColAdd; // For animation
  XColor24    xInfoColor;        // The info color
  char        acBuff[128];       // The information string buffor
  EKeySymbol  eKS;               // The key symbol

  // UTF8 sample text
  // NOTE: - The characters between << and >> are Japanese characters
  //       - One will need to install the needed fonts in order to display them correctly
  const char *utf8_cstr = "UTF-8 sample text << ぁぃぅぇぉ ァィゥェォ 一二三 >>";

  // Test the CUString class
  const CUString custr1(utf8_cstr, CUString::ESCE_Utf8); 
  const CUString custr2(custr1); 
        CUString custr3; custr3 = custr2;
  const string  &utf8Text = custr3.ToUtf8();

  // Initialization
  VPrintCPUCaps(true);
  if(!InitializeSystem()) return(1);

  // Tell the user to be patient
  ShowInitSystemPleaseWaitMessage(csVidEnv);

  // Load medias
  LoadImages();
  LoadAudios();

  // Create scene
  xInfoColor = csVidEnv.CreateColor(255, 255, 255);
  csMat.SetToIdentityMatrix();
  csMat.Translate(0, 20, 150);
  csMat.Rotate(100, 0, 0);
  cs3DEnv.SetFogState(8, 16, 12, 13);
  cs3DEnv.SetDefaultOmniLightPosition(40, 200, 150);
  cs3DEnv.SetDefaultOmniLightColor(31, 31, 16);
  cs3DEnv.SetZClipping(200, 500);
  CreateScene();

  // Play background audio
  csSBBack.Play(true);

  // Initialize variables
  fALColAdd = 0.5;
  fALCol    = 8.0;
  bGouraudShaded = true;
  bBackImg       = true;
  bTransparent   = false;
  bFog           = true;
  un8TexType     = 2;
  un8ADir        = 0;
  bAlwaysOnTop   = false;
  bFullscreen    = false;

  // Create font
  TFont *ptFont = NULL;
  try {
    ptFont = csVidEnv.CreateFont(CFontDesc("times",
                                           "12/11/10",
                                           CFontDesc::Pitch::Any,
                                           CFontDesc::Weight::Regular,
                                           CFontDesc::Slant::Italic,
                                           CFontDesc::Underline::None,
                                           "iso8859"));
  }
  catch(...) {
    ptFont = csVidEnv.DefaultSmallFont();
  }
  csVidEnv.SetFont(ptFont);

  // Create UTF-8 font
  TFontSet *ptFontSet = NULL;
  try {
    ptFontSet = csVidEnv.CreateUtf8Font(CFontDesc("*", "15/16/17/18/19"));
  }
  catch(...) {
    ptFontSet = csVidEnv.DefaultUtf8SmallFont();
  }
  csVidEnv.SetUtf8Font(ptFontSet);

  // Main loop
  csVidEnv.AutoRepeatOff();
  csFPSCtrl.SaveStartTicks();
  VInfo(_NFL_, "Main loop is starting ...\n");
  while(1) {
    // Frame rate control
    if(csFPSCtrl.PumpEventAndDelay(csVidEnv)) continue;
    if(csFPSCtrl.AdjustDelay()) snprintf(acBuff, sizeof(acBuff),
                                         "%05.1f fps, %06d delays",
                                         csFPSCtrl.GetFPS(), csFPSCtrl.GetDelayFactor());

    // Update everything (the application is active)
    if(csVidEnv.IsActive()) {
      // Check event
      if(csVidEnv.CheckEvent()) {
        // Keyboard event
        if(csVidEnv.EventKeyPress()) {
          char kcb[32];
          if(csVidEnv.GetKeyString(kcb, sizeof(kcb)) && kcb[0])
            VPrint("KeyPress   : %05u (%05lu) [%c]\n", csVidEnv.GetKeySym(), csVidEnv.GetRawKeyCode(), kcb[0]);
          else
            VPrint("KeyPress   : %05u (%05lu) [ ]\n", csVidEnv.GetKeySym(), csVidEnv.GetRawKeyCode());
        }
        else if(csVidEnv.EventKeyRelease()) {
          eKS = csVidEnv.GetKeySym();
               if(eKS == 'b') bBackImg = !bBackImg;
          else if(eKS == 's') bGouraudShaded = !bGouraudShaded;
          else if(eKS == 't') {
            un8TexType++;
            if(un8TexType > 3) un8TexType = 0;
          }
          else if(eKS == 'f') bFog = !bFog;
          else if(eKS == 'n') bTransparent = !bTransparent;
          else if(eKS == 'a') {
            un8ADir = (un8ADir == 0) ? 1 : 0;
            csArrowCtl.SelectAnimationSet(un8ADir);
          }
          else if(eKS == 'z') csSBShot.Play();
          else if(eKS == 'o') {
            if(!bFullscreen){
              bAlwaysOnTop = !bAlwaysOnTop;
              csVidEnv.SetAlwaysOnTop(bAlwaysOnTop);
            }
          }
          else if(eKS == 'w') {
            bFullscreen = !bFullscreen;
                 if(bFullscreen && bAlwaysOnTop)    csVidEnv.SetAlwaysOnTop(false);
            else if((!bFullscreen) && bAlwaysOnTop) csVidEnv.SetAlwaysOnTop(true);
            csVidEnv.SetFullscreen(bFullscreen);
          }
          else if(eKS == 'x') break;

          char kcb[32];
          if(csVidEnv.GetKeyString(kcb, sizeof(kcb)) && kcb[0])
            VPrint("KeyRelease : %05u (%05lu) [%c]\n", eKS, csVidEnv.GetRawKeyCode(), kcb[0]);
          else
            VPrint("KeyRelease : %05u (%05lu) [ ]\n", eKS, csVidEnv.GetRawKeyCode());
        }
        // Mouse event
        else if(csVidEnv.EventButtonPress())
          VPrint("BtnPress   : %d\n", csVidEnv.GetButtonSym());
        else if(csVidEnv.EventButtonRelease()) {
          VPrint("BtnRelease : %d\n", csVidEnv.GetButtonSym());
        }
        // Other event
        else if(csVidEnv.EventCloseWindow())
          break;
      }

      // Don't update till the delay counter reach zero
      if(csFPSCtrl.GetDelayCounter()) continue;

      // Update scene
      fALCol += fALColAdd;
      if(fALCol > 32) {
        fALCol = 32;
        fALColAdd = -fALColAdd;
      }
      if(fALCol < 8) {
        fALCol = 8;
        fALColAdd = -fALColAdd;
      }
      cs3DEnv.SetDefaultAmbientLightColor(uint8_t(fALCol), uint8_t(fALCol), uint8_t(fALCol));
      if(bGouraudShaded)
        csWorld.GSResetAmbientLight();
      else
        csWorld.FSResetAmbientLight();
      csTreeSpr.ResetAmbientLight();
      csLightImg.StepFrame(5);
      csArrowCtl.SelectNextASFrameDly(5, true);

      // Render scene
      if(bBackImg)
        cs3DEnv.ClearRenderBuffer(csBackImg);
      else
        cs3DEnv.ClearRenderBuffer(V_RGB(0, 4, 0));
      cs3DEnv.ClearZBuffer();
      RenderPanels();
      RenderSprites();
#ifdef __HIGH_RESOLUTION__
      csBF.PutText(10, 415, "BF : Regular",     VColor::MediumAquamarine, EFS_Regular);
      csBF.PutText(10, 430, "BF : Italic",      VColor::MediumAquamarine, EFS_Italic);
      csBF.PutText(10, 445, "BF : Bold",        VColor::MediumAquamarine, EFS_Bold);
      csBF.PutText(10, 460, "BF : Bold Italic", VColor::MediumAquamarine, EFS_BoldItalic);
#else
      csBF.PutText(10, 175, "BF : Regular",     VColor::MediumAquamarine, EFS_Regular);
      csBF.PutText(10, 190, "BF : Italic",      VColor::MediumAquamarine, EFS_Italic);
      csBF.PutText(10, 205, "BF : Bold",        VColor::MediumAquamarine, EFS_Bold);
      csBF.PutText(10, 220, "BF : Bold Italic", VColor::MediumAquamarine, EFS_BoldItalic);
#endif

      // Update display
      csVidEnv.PutRenderImageOnDisplay();
      csVidEnv.SetForegroundColor(xInfoColor);
      csVidEnv.DrawText(10, 20, acBuff);
           if(bBackImg)        csVidEnv.DrawText(10, 50, "Press 'B' to change background image->solid fill");
      else                     csVidEnv.DrawText(10, 50, "Press 'B' to change solid fill->background image");
           if(bGouraudShaded)  csVidEnv.DrawText(10, 65, "Press 'S' to change GS->FS");
      else                     csVidEnv.DrawText(10, 65, "Press 'S' to change FS->GS");
           if(un8TexType == 0) csVidEnv.DrawText(10, 80, "Press 'T' to use solid color");
      else if(un8TexType == 1) csVidEnv.DrawText(10, 80, "Press 'T' to use texture");
      else if(un8TexType == 2) csVidEnv.DrawText(10, 80, "Press 'T' to use texture without any shading");
      else                     csVidEnv.DrawText(10, 80, "Press 'T' to use light color only");
           if(bFog)            csVidEnv.DrawText(10, 95, "Press 'F' to remove fog");
      else                     csVidEnv.DrawText(10, 95, "Press 'F' to use fog");
           if(bTransparent)    csVidEnv.DrawText(10, 110, "Press 'N' to use solid sprite");
      else                     csVidEnv.DrawText(10, 110, "Press 'N' to use transparent sprite");
      csVidEnv.DrawText(10, 125, "Press 'A' to reverse the arrow direction");
      csVidEnv.DrawText(10, 155, "Press 'Z' to play a one-shot sound buffer");
      if(bFullscreen)
        csVidEnv.DrawText(10, 200, "Press 'W' to switch to windowed mode");
      else {
        if(bAlwaysOnTop) csVidEnv.DrawText(10, 185, "Press 'O' to disable always on top");
        else             csVidEnv.DrawText(10, 185, "Press 'O' to enable always on top");
        csVidEnv.DrawText(10, 200, "Press 'W' to switch to full screen mode");
      }
      csVidEnv.DrawText(10, 230, "Press 'X' to exit ...");
      csVidEnv.DrawUtf8Text(10, 280, utf8Text);
      csVidEnv.UpdateDisplay();

      // Update audio
      csAudioEnv.UpdateBuffer();
    }
    // Update everything (the application is inactive)
    else {
      // Make sure that we check for event eventough we won't use it
      // (if we don't do this, the event pumping may get blocked)
      csVidEnv.CheckEvent();

      // Update display (only of the application window is visible)
      if(csVidEnv.IsVisible()) {
        csVidEnv.PutRenderImageOnDisplay();
        csVidEnv.SetForegroundColor(xInfoColor);
        csVidEnv.DrawText(10, 20, acBuff);
        csVidEnv.UseDefaultBigFont();
        csVidEnv.DrawText(0, 150, "Inactive State", ETA_CenterScreen);
        csVidEnv.SetFont(ptFont);
        csVidEnv.UpdateDisplay();
      }
    }
  }

  // Closing down
  csFPSCtrl.SaveEndTicks();
  csVidEnv.FreeFont(ptFont);
  csVidEnv.FreeUtf8Font(ptFontSet);
  csAudioEnv.DeactivateDevice();

  // Calculate & display the avegare frame rate
  snprintf(acBuff, sizeof(acBuff), "Average frame rate = %.2f fps\n\nClick \"OK\" to exit ...", csFPSCtrl.GetAvgFPS());
  csVidEnv.ApplyBlackScreen();
  csVidEnv.MessageBox("End Program", acBuff);
  VInfo(_NFL_, "Program ended, average frame rate = %.2f FPS", csFPSCtrl.GetAvgFPS());

  // OK !
  return(0);
}


// Initialize system
bool InitializeSystem()
{
  // Temporary variables
#ifdef __HIGH_RESOLUTION__
  uint32_t un32W = 640, un32H = 480, un32D = 1;
#else
  uint32_t un32W = 640, un32H = 480, un32D = 2;
#endif
  char     acBuff[PATH_MAX];

  // Initialize message catalog manually (since we are not using the high level API)
  try { CBasicAppObjects::InitializeI18N(); }
  catch(...) {}

  // Initialize main stuff
  csVidEnv.CreateMainWindow(un32W, un32H, "VortexGE Test Application", "../media/image/wave.xpm");
  csVidEnv.CreateRenderImage(un32W, un32H);
  csVidEnv.CreateRenderBuffer(un32W / un32D, un32H / un32D);
  cs3DEnv.Initialize(&csVidEnv);
  csAudioEnv.QueryDeviceName(acBuff, sizeof(acBuff));
  for(;;) {
    char          acTmp[1024];
    EMsgBoxRetVal eRet;

    bool bOpened = false;
    try {
      csAudioEnv.OpenDevice(acBuff);
      bOpened = true;
    }
    catch(...) {}
    if(bOpened) {
      csAudioEnv.InitializeDevice();
      csAudioEnv.ActivateDevice();
      break;
    }

    const char *t0 = "CAudioEnv failed to open audio device \"%s\", %s%s%s%s%s";
    const char *t1 = "the audio device may be in use.\n\n";
    const char *t2 = "Terminate all other multimedia applications or wait for about 30-60 seconds or\n";
    const char *t3 = "until a few minutes till any sound server, such as \"artsd\" or \"esd\") suspends itself.\n\n";
    const char *t4 = "Click abort to terminate this program ...\n";
    const char *t5 = "Click retry to try again ...\nClick ignore to continue without audio ...";
    snprintf(acTmp, sizeof(acTmp),
             t0, acBuff, t1, t2, t3, t4, t5);
    eRet = csVidEnv.MessageBox("Couldn't open Audio Device", acTmp, EMBI_Warning, EMBB_AbortRetryIgnore, EMBS_Medium);
    if(eRet == EMBRV_Abort) return(false);
    if(eRet == EMBRV_Ignore) {
      VWarning(_NFL_, "Audio is disabled");
      break;
    }
  }

  // OK !
  return(true);
}

// Load images
void LoadImages()
{
  // Load background image
  try {
    csImgBuild.LoadVIB("../.mediaCaches/back2.vib");
  }
  catch(...) {
    try {
      csImgBuild.LoadJPG("../media/image/back2.jpg", true);
    }
    catch(...) {
      csImgBuild.CreateNoiseImage(256, 256, 8, 24, 8, 3, 3, 3);
    }
    csImgBuild.SaveVIB("../.mediaCaches/back2.vib");
  }
  csImgBuild.ResizeToFitBackground();
  csImgBuild.AdjustBrightness(0.25, 0);
  csImgBuild.SmearEffect(5);
  csImgBuild.TransferToImage(csBackImg);

  // Load texture image
  try {
    csImgBuild.LoadVIB("../.mediaCaches/tex2.vib");
  }
  catch(...) {
    try {
      csImgBuild.LoadJPG("../media/image/tex2.jpg", true);
      csImgBuild.Resize(128, 128);
    }
    catch(...) {
      csImgBuild.CreateNoiseImage(128, 128, 14, 14, 20, 6, 6, 2);
      csImgBuild.OilPaintingEffect(4);
    }
    // Darken the horizontal borders
    TUN8RGB *lines[] = { csImgBuild.GetLineBuffer(0),
                         csImgBuild.GetLineBuffer(1),
                         csImgBuild.GetLineBuffer(2),
                         csImgBuild.GetLineBuffer(3),
                         csImgBuild.GetLineBuffer(csImgBuild.GetHeight() - 4),
                         csImgBuild.GetLineBuffer(csImgBuild.GetHeight() - 3),
                         csImgBuild.GetLineBuffer(csImgBuild.GetHeight() - 2),
                         csImgBuild.GetLineBuffer(csImgBuild.GetHeight() - 1) };
    for(uint32_t c = 0; c < elemsize(lines); ++c) {
      for(uint32_t x = 0; x < csImgBuild.GetWidth(); ++x) {
        (lines[c] + x)->un8R = uint8_t((lines[c] + x)->un8R * 0.75f);
        (lines[c] + x)->un8G = uint8_t((lines[c] + x)->un8G * 0.75f);
        (lines[c] + x)->un8B = uint8_t((lines[c] + x)->un8B * 0.75f);
      }
    }
    // Darken the vertical borders
    const uint32_t ofs[] = { 0, 1, 2, 3,
                            csImgBuild.GetWidth() - 4,
                            csImgBuild.GetWidth() - 3,
                            csImgBuild.GetWidth() - 2,
                            csImgBuild.GetWidth() - 1 };
    for(uint32_t y = 4; y < (csImgBuild.GetHeight() - 4); ++y) {
      TUN8RGB *line = csImgBuild.GetLineBuffer(y);
      for(uint32_t c = 0; c < elemsize(ofs); ++c) {
        (line + ofs[c])->un8R = uint8_t((line + ofs[c])->un8R * 0.75f);
        (line + ofs[c])->un8G = uint8_t((line + ofs[c])->un8G * 0.75f);
        (line + ofs[c])->un8B = uint8_t((line + ofs[c])->un8B * 0.75f);
      }
    }
    // Blur and save the image
    csImgBuild.BlurEffect(1);
    csImgBuild.SaveVIB("../.mediaCaches/tex2.vib");
  }
  csImgBuild.TransferToTexture(csTex);

  // Load tree sprite image
  try {
    csImgBuild.LoadVIB("../.mediaCaches/tree1.vib");
  }
  catch(...) {
    csImgBuild.LoadBMP("../media/image/tree1.bmp", true);
    csImgBuild.SaveVIB("../.mediaCaches/tree1.vib");
  }
  csImgBuild.RoughResize(250, 0);
  csImgBuild.TransferToImage(csTreeImg);

  // Load the light sprite image
  try {
    csImgBuild.LoadVIB("../.mediaCaches/sparklight.vib");
  }
  catch(...) {
    csImgBuild.LoadBMP("../media/image/sparklight.bmp", true);
    csImgBuild.SaveVIB("../.mediaCaches/sparklight.vib");
  }
  csImgBuild.Resize(320*4, 32*4);
  csImgBuild.AdjustBrightness(0.5, 0, true, true, false);
  csImgBuild.TransferToAnimatedImage(csLightImg, 1, 10);

  // Load the arrow sprite image
  try {
    csImgBuild.LoadVIB("../.mediaCaches/arrow.vib");
  }
  catch(...) {
    csImgBuild.LoadBMP("../media/image/arrow.bmp", true);
    csImgBuild.SaveVIB("../.mediaCaches/arrow.vib");
  }
  csImgBuild.TransferToAnimatedImage(csArrowImg, 2, 5);
  csArrowCtl.SetImageSource(&csArrowImg);
  try {
    csArrowCtl.LoadVAC("../.mediaCaches/arrow.vac");
  }
  catch(...) {
    csArrowCtl.AddAnimationSets(2);
    csArrowCtl.AddASFrame(0, 0);
    csArrowCtl.AddASFrame(0, 1);
    csArrowCtl.AddASFrame(0, 2);
    csArrowCtl.AddASFrame(0, 3);
    csArrowCtl.AddASFrame(0, 4);
    csArrowCtl.AddASFrame(1, 5);
    csArrowCtl.AddASFrame(1, 6);
    csArrowCtl.AddASFrame(1, 7);
    csArrowCtl.AddASFrame(1, 8);
    csArrowCtl.AddASFrame(1, 9);
    csArrowCtl.SaveVAC("../.mediaCaches/arrow.vac");
  }

  // Free image
  csImgBuild.FreeImage();

  // Load bitmap font
  try {
    csBF.LoadVBF("../.mediaCaches/courier-12pt.vbf");
  }
  catch(...) {
    csBF.LoadBMP("../media/image/font/courier-12pt.bmp");
    csBF.SaveVBF("../.mediaCaches/courier-12pt.vbf");
  }
}

// Load audios
void LoadAudios()
{
  // Check if disabled
  if(!csAudioEnv.IsAvailable()) return;

  // Load background audios
  try {
    csSBBack.LoadVSW("../.mediaCaches/backloop.vsw");
  }
  catch(...) {
    //csSBBack.LoadRAW("../media/raw/puck-S16Ls22.raw", ERF_S16LEStereo, 22050);
    csSBBack.LoadRAW("../media/raw/puck-S16Ls48.raw", ERF_S16LEStereo, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-S16Bs48.raw", ERF_S16BEStereo, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-U16Ls48.raw", ERF_U16LEStereo, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-U16Bs48.raw", ERF_U16BEStereo, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-S16Lm48.raw", ERF_S16LEMono, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-S16Bm48.raw", ERF_S16BEMono, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-U16Lm48.raw", ERF_U16LEMono, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-U16Bm48.raw", ERF_U16BEMono, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-S8s48.raw", ERF_S8Stereo, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-U8s48.raw", ERF_U8Stereo, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-S8m48.raw", ERF_S8Mono, 48000);
    //csSBBack.LoadRAW("../media/raw/puck-U8m48.raw", ERF_U8Mono, 48000);
    csSBBack.SaveVSW("../.mediaCaches/backloop.vsw");
  }
  csSBBack.AdjustAmplitude(0.2f, 0.2f);

  // Load foreground audios
  try {
    csSBShot.LoadVSW("../.mediaCaches/foreshot.vsw");
  }
  catch(...) {
    csSBShot.LoadWAV("../media/wav/jump.wav");
    csSBShot.SaveVSW("../.mediaCaches/foreshot.vsw");
  }
  csSBShot.AdjustAmplitude(0.4f, 0.4f);
}

// Create scene
void CreateScene()
{
  // Temporary variables
  float afV[] = { -60,  -50, 100,   -20,  -50, 100,   20,  -50, 100,   60,  -50, 100,
                  -60,    0, 140,   -20,    0, 140,   20,    0, 140,   60,    0, 140,
                  -60,   50, 180,   -20,   50, 180,   20,   50, 180,   60,   50, 180,
                  -60,  100, 220,   -20,  100, 220,   20,  100, 220,   60,  100, 220,
                  -60,    0, 100,   -60,   50, 140,  -60,  100, 180,  -60,  150, 220,
                   60,    0, 100,    60,   50, 140,   60,  100, 180,   60,  150, 220,
                  -20,  150, 220,    20,  150, 220 };
  uint32_t aun32P[] = { 0,   4,  5,  1,    1,  5,  6,  2,    2,  6,  7,  3,
                        4,   8,  9,  5,    5,  9, 10,  6,    7,  6, 10, 11,
                        9,   8, 12, 13,    9, 13, 14, 10,   10, 14, 15, 11,
                        0,  16, 17,  4,    4, 17, 18,  8,    8, 18, 19, 12,
                        3,   7, 21, 20,    7, 11, 22, 21,   11, 15, 23, 22,
                       12,  19, 24, 13,   13, 24, 25, 14,   14, 25, 23, 15 };
  float afSPos1[] = { -30, -15, 120,   30, 85, 200 };
  float afSPos2[] = { 100, 100, 120 };

  // Create scene
  try {
    csWorld.LoadVPD("../.mediaCaches/deck.vpd");
    csWorld.SetTexture(&csTex);
  }
  catch(...) {
    csWorld.InitFromMemory(afV, 26, aun32P, 18);
    csWorld.SetColor(24, 31, 24);
    csWorld.SetTexture(&csTex);
    csWorld.CalcDefTexCoords();
    csWorld.SaveVPD("../.mediaCaches/deck.vpd");
  }
  csTreeSpr.Initialize(afSPos1, 2, &csTreeImg);
  csLightSpr.Initialize(afSPos2, 1, &csLightImg, EP_CenterCenter);

  //csWorld.SaveVPD("../.mediaCaches/deck.vpd");
  //csWorld.__SaveVPD111("../.mediaCaches/deck.vpd");
  //csWorld.__SaveVPD110("../.mediaCaches/deck.vpd");
  //csWorld.__SaveVPD100("../.mediaCaches/deck.vpd");

  //csWorld.LoadVPD("../.mediaCaches/deck.vpd");
  //csWorld.__LoadVPD111("../.mediaCaches/deck.vpd");
  //csWorld.__LoadVPD110("../.mediaCaches/deck.vpd");
  //csWorld.__LoadVPD100("../.mediaCaches/deck.vpd");
}


// Render the panels
void RenderPanels()
{
  if(bFog) {
    if(un8TexType == 0) {
      if(bGouraudShaded)
        csWorld.FogUGRender(csMat);
      else
        csWorld.FogUFRender(csMat);
    }
    else if(un8TexType == 1) {
      if(bGouraudShaded)
        csWorld.FogCGRender(csMat);
      else
        csWorld.FogCFRender(csMat);
    }
    else if(un8TexType == 2) {
      if(bGouraudShaded)
        csWorld.FogTGRender(csMat);
      else
        csWorld.FogTFRender(csMat);
    }
    else {
      csWorld.FogTNRender(csMat);
    }
  }
  else {
    if(un8TexType == 0) {
      if(bGouraudShaded)
        csWorld.UGRender(csMat);
      else
        csWorld.UFRender(csMat);
    }
    else if(un8TexType == 1) {
      if(bGouraudShaded)
        csWorld.CGRender(csMat);
      else
        csWorld.CFRender(csMat);
    }
    else if(un8TexType == 2) {
      if(bGouraudShaded)
        csWorld.TGRender(csMat);
      else
        csWorld.TFRender(csMat);
    }
    else {
      csWorld.TNRender(csMat);
    }
  }
}

// Render the sprites
void RenderSprites()
{
  // Render the sprites
  if(bFog) {
    if(bTransparent) {
      if(un8TexType <= 2)
        csTreeSpr.FogTFRender(csMat);
      else
        csTreeSpr.FogTNRender(csMat);
    }
    else {
      if(un8TexType <= 2)
        csTreeSpr.FogSFRender(csMat);
      else
        csTreeSpr.FogSNRender(csMat);
    }
  }
  else {
    if(bTransparent) {
      if(un8TexType <= 2)
        csTreeSpr.TFRender(csMat);
      else
        csTreeSpr.TNRender(csMat);
    }
    else {
      if(un8TexType <= 2)
        csTreeSpr.SFRender(csMat);
      else
        csTreeSpr.SNRender(csMat);
    }
  }
  csLightSpr.ARender(csMat);
#ifdef __HIGH_RESOLUTION__
  cs3DEnv.PutImage(csArrowImg, 640-107, 480-45, VColor::Black);
#else
  cs3DEnv.PutImage(csArrowImg, 320-107, 240-45, VColor::Black);
#endif
}
