// ############################################################################################
//  VLGame-AppClass-EH.cc
//  A demo application for VortexGE (implementation of VLGame's application class)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VLGame's headers
#include "VLGame-AppClass.h"


//
// VLGApplication implementation
//
void VLGApplication::EventKeyPress(EKeySymbol eKeySym)
{
       if(eKeySym == '\\'     ) iVideoEnv().SetFullscreen(!iVideoEnv().IsFullscreen());
  else if(eKeySym == VKS_ESC  ) EventCloseWindow();
  else if(eKeySym == VKS_UP   ) keyMap.SetArrows(true , false, false, false);
  else if(eKeySym == VKS_DOWN ) keyMap.SetArrows(false, true , false, false);
  else if(eKeySym == VKS_LEFT ) keyMap.SetArrows(false, false, true , false);
  else if(eKeySym == VKS_RIGHT) keyMap.SetArrows(false, false, false, true );
  else if(eKeySym == ' '      ) {
    if(uscStartRoom->InDoorAccessArea(player.GetPosition(), player.GetRadius())) {
      if(uscStartRoom->DoorIsOpen()) uscStartRoom->CloseDoor();
      else uscStartRoom->OpenDoor();
      if(!AudioIsDisabled() && !sbDoor->IsPlaying()) sbDoor->Play();
    }
  }
}

void VLGApplication::EventKeyRelease(EKeySymbol eKeySym)
{
       if(eKeySym == VKS_UP   ) keyMap.up    = false;
  else if(eKeySym == VKS_DOWN ) keyMap.down  = false;
  else if(eKeySym == VKS_LEFT ) keyMap.left  = false;
  else if(eKeySym == VKS_RIGHT) keyMap.right = false;
}

void VLGApplication::EventUpdateFrame(bool bWindowIsActive)
{
  // Is the application is not active ?
  if(!bWindowIsActive) {
    // Make sure that the audio thread is not suspended (just in case that
    // EventWindowDeactivated() was not dispatched correctly)
    if(!AudioIsDisabled()) audioThread.Suspend();
    // Display simple information
    iVideoEnv().PutRenderImageOnDisplay();
    iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
    iVideoEnv().UseDefaultBigFont();
    iVideoEnv().DrawText(0, 150, "Inactive State", ETA_CenterScreen);
    iVideoEnv().UseDefaultSmallFont();
    iVideoEnv().UpdateDisplay();
    // Nothing else to do, so just exit
    return;
  }

  // Make sure that the audio thread is running (just in case that
  // EventWindowActivated() was not dispatched correctly)
  if(!AudioIsDisabled()) audioThread.Resume();

  // Update player (and the follow-camera positions)
  if(player.Move(*uscStartRoom, keyMap.up, keyMap.down, keyMap.left, keyMap.right)) {
    if(!AudioIsDisabled() && !sbWalk->IsPlaying()) sbWalk->Play();
  }

  // Update scene
  uscStartRoom->UpdateScene();

  // Clear scene
  i3DEnv().ClearRenderBuffer(VColor::Black);
  i3DEnv().ClearZBuffer();

  // Render scene (all solid objects)
  uscStartRoom->RenderAllSolidObjects(player.FCamMatrix());
  player.iSprite2D().FogSNRender(player.FCamMatrix());

  // Render scene (all transparent objects)
  uscStartRoom->RenderAllTransparentObjects(player.FCamMatrix());

  // Update the player position information string
  char playerPosition[256];
  snprintf(playerPosition, sizeof(playerPosition),
           "Player is at (%+07.1f, %+07.1f, %+07.1f)",
           player.GetXPos(), player.GetYPos(), player.GetZPos());

  // Print some text and update display
  iVideoEnv().PutRenderImageOnDisplay();
  iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
  iVideoEnv().DrawText(10,  20, "Use arrows to move ...");
  iVideoEnv().DrawText(10,  35, "Use space to interact ...");
  iVideoEnv().DrawText( 0, 230, "[This demo application is not finished yet]", ETA_CenterScreen);
  iVideoEnv().DrawText(10, 438, playerPosition);
  iVideoEnv().DrawText(10, 468, GetFPSInfoString());
  iVideoEnv().DrawText( 0, 468, "Press escape to exit ...", ETA_CenterScreen);
  iVideoEnv().UpdateDisplay();
  
}


void VLGApplication::EventWindowActivated()
{
  if(!AudioIsDisabled()) audioThread.Resume();
}

void VLGApplication::EventWindowDeactivated()
{
  if(!AudioIsDisabled()) audioThread.Suspend();
}

void VLGApplication::EventCloseWindow()
{
  EMsgBoxRetVal eRet;
  eRet = iVideoEnv().MessageBox("Confirmation", "Do you really want to exit ?", EMBI_Question, EMBB_YesNo, EMBS_Small);
  if(eRet == EMBRV_Yes) {
    luaManager.SaveWorkingDirectory(VLGApplication::SaveGameFileName);
    VLGApplicationBase::EventCloseWindow();
  }
}
