// ############################################################################################
//  VLGame-Entities.cc
//  A demo application for VortexGE (implementation of VLGame's entity classes)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VLGame's headers
#include "VLGame-Entities.h"


//
// VLGPlayer implementation
//
const float VLGPlayer::CameraYPos =  350.0f;
const float VLGPlayer::CameraZPos = -100.0f;


VLGPlayer::VLGPlayer(const TInit_BasicCharSpr2D &init)
: VIC_CBasicCharacterSprite2D(init)
{
}

void VLGPlayer::Initialize(const CAnimatedImage_MP &aimg, const TFloatXYZ &initPos)
{
  _pos = initPos;
  ReinitializeSprite2D(aimg, NULL, &initPos);

  TFloatXY min, max;
  iSprite2D().GetEquivalentGeometry(min, max);
  _rad.x = max.x - min.x;
  _rad.y = max.y - min.y;
  _rad.z = 0;

  // Initialize camera
  _fCamTarget[0] = _pos.x; _fCamPos[0] = _fCamTarget[0];
  _fCamTarget[1] = 0;      _fCamPos[1] = CameraYPos;
  _fCamTarget[2] = _pos.z; _fCamPos[2] = _fCamTarget[2] + CameraZPos;
  _fCam.SetToLookAtMatrix(_fCamPos, _fCamTarget);
}

bool VLGPlayer::MoveLeft(const CUnitSceneObjects &usc, float amount)
{ 
  iSpriteAImage()->SetFrame(2);
  if(usc.PassableRegion(TFloatXYZ(_pos.x - amount, _pos.y, _pos.z), _rad)) {
    AdjustXPos(-amount); 
    iSprite2D().SetPosition(0, _pos.x, _pos.y, _pos.z);
    FCamUpdate();
    return(true);
  }
  return(false);
}

bool VLGPlayer::MoveRight(const CUnitSceneObjects &usc, float amount)
{ 
  iSpriteAImage()->SetFrame(1);
  if(usc.PassableRegion(TFloatXYZ(_pos.x + amount, _pos.y, _pos.z), _rad)) {
    AdjustXPos(amount); 
    iSprite2D().SetPosition(0, _pos.x, _pos.y, _pos.z);
    FCamUpdate();
    return(true);
  }
  return(false);
}

bool VLGPlayer::MoveUp(const CUnitSceneObjects &usc, float amount)
{ 
  iSpriteAImage()->SetFrame(0);
  if(usc.PassableRegion(TFloatXYZ(_pos.x, _pos.y, _pos.z + amount), _rad)) {
    AdjustZPos(amount); 
    iSprite2D().SetPosition(0, _pos.x, _pos.y, _pos.z);
    FCamUpdate();
    return(true);
  }
  return(false);
}

bool VLGPlayer::MoveDown(const CUnitSceneObjects &usc, float amount)
{ 
  iSpriteAImage()->SetFrame(3);
  if(usc.PassableRegion(TFloatXYZ(_pos.x, _pos.y, _pos.z - amount), _rad)) {
    AdjustZPos(-amount); 
    iSprite2D().SetPosition(0, _pos.x, _pos.y, _pos.z);
    FCamUpdate();
    return(true);
  }
  return(false);
}

bool VLGPlayer::Move(const CUnitSceneObjects &usc, bool up, bool down, bool left, bool right)
{
       if(up   ) return(MoveUp   (usc));
  else if(down ) return(MoveDown (usc));
  else if(left ) return(MoveLeft (usc));
  else if(right) return(MoveRight(usc));

  return(false);
}

void VLGPlayer::FCamUpdate()
{
  _fCamTarget[0] = _pos.x; _fCamPos[0] = _fCamTarget[0];
  _fCamTarget[2] = _pos.z; _fCamPos[2] = _fCamTarget[2] + CameraZPos;
  _fCam.SetToLookAtMatrix(_fCamPos, _fCamTarget);
}


//
// VLGUnitScene_StartRoom implementation
//
VLGUnitScene_StartRoom::VLGUnitScene_StartRoom(CLuaManager        &luaManager,
                                               CArchiveDataSource &mediaPackage,
                                               CMediaPool      &mediaPool,
                                               C3DEnv             &i3DEnv)
: _luaManager  (luaManager),
  _mediaPackage(mediaPackage),
  _mediaPool   (mediaPool),
  _i3DEnv      (i3DEnv),
  _fireAnim    (TInit_BasicCharSpr2D("fire animation", NULL, NULL, NULL, &_luaManager)),
  _doorOpened  (false),
  _doorOrder   (0),
  _doorCnt     (0)
{
  _i3DEnv.SetAsTheActiveEnv();

  // Initialize "room"
#if 0
  CPlane3DBuilder roomPanels(TFloatXYZ(20.0f, 20.0f, 20.0f));
  roomPanels.GeneratePlane("land", TFloatXYZ(-800.0f, 0.0f, 1500.0f), 80, 80, EPD_0P0);
  roomPanels["land"].RandomlyVaryHeights(25);
#else
  CPlane3DBuilder roomPanels;
  try {
    roomPanels.LoadVPB("../.mediaCaches/room.vpb");
  }
  catch(...) {
    roomPanels.SetGridSize(TFloatXYZ(100.0f, 100.0f, 100.0f));
    roomPanels.CreateSimpleOpenRoom(TFloatXYZ(-300.0f, 0.0f, 500.0f),
                                    TInt32XYZ(6, 1, 4));
    roomPanels.SaveVPB("../.mediaCaches/room.vpb");
  }
#endif
  roomPanels.GeneratePanel3DBuffer(_room, _mediaPool.GetTexture("Texture Room"));
  roomPanels.GetBoundingBox(_areaFloor[0], _areaFloor[1]);
  //VPrint("%.2f,%.2f,%.2f %.2f,%.2f,%.2f\n",
  //       _areaFloor[0].x, _areaFloor[0].y, _areaFloor[0].z,
  //       _areaFloor[1].x, _areaFloor[1].y, _areaFloor[1].z);
  _areaFloor[0].x += 100;
  _areaFloor[1].x -= 100;
  _areaFloor[1].z -= 110;

  // Initialize "door"
  float afMDoor[]       = { -225,100,100, -175,100,100, -250,75,100, -150,75,100,
                            -250,  0,100, -225,  0,100, -175, 0,100, -150, 0,100,
                            -225,100, 90, -175,100, 90, -250,75, 90, -150,75, 90,
                            -250,  0, 90, -225,  0, 90, -175, 0, 90, -150, 0, 90 };
  uint32_t aun32MDoor[] = { 5, 0, 2, 4,  6,1,0, 5,  7, 3, 1, 6,
                           10, 2, 0, 8,  8,0,1, 9,  9, 1, 3,11,
                            4, 2,10,12,            15,11, 3, 7,
                           12,10, 8,13, 13,8,9,14, 14, 9,11,15 };
  _door.InitFromMemory(afMDoor, 16, aun32MDoor, 11);
  _door.SetTexture(_mediaPool.GetTexture("Texture Door"));
  _door.CalcDefTexCoords();
  _door.GetBoundingBox(_areaDoor[0], _areaDoor[1]);

  // Initialize "path"
  float afMPath[]       = { -250,  0,110, -150,  0,110,
                            -250,-10, 60, -150,-10, 60,
                            -260,-20, 10, -160,-20, 10,
                            -280,-30,-40, -180,-30,-40 };
  uint32_t aun32MPath[] = { 2,0,1,3,  4,2,3,5,  6,4,5,7 };
  _path.InitFromMemory(afMPath, 8, aun32MPath, 3);
  _path.SetTexture(_mediaPool.GetTexture("Texture Room"));
  _path.CalcDefTexCoords();
  _path.GetBoundingBox(_areaPath[0], _areaPath[1]);
  _areaPath[0].x += 30;
  _areaPath[0].z += 80;

  // Initialize "soil"
  float afMSoil[]       = { -700,-28, 800,  700,-28, 800,
                            -700,-28,-300,  700,-28,-300 };
  uint32_t aun32MSoil[] = { 2,0,1,3,};
  _soil.InitFromMemory(afMSoil, 4, aun32MSoil, 1);
  _soil.SetColor(24, 24, 31);

  // Initialize "fire animation"
  vector<TFloatXYZ> firePos(3);
  firePos[0] = TFloatXYZ(-200, 26, 350);
  firePos[1] = TFloatXYZ(   0, 26, 350);
  firePos[2] = TFloatXYZ(-100, 26, 450);

  _fireAnim.ReinitializeSprite2D(CAnimatedImage_MP(_mediaPool, "Animated Fire"), NULL, &firePos);
}

VLGUnitScene_StartRoom::~VLGUnitScene_StartRoom() throw()
{
  //! \todo Find a better way to do this automatically !!!
  _mediaPool.ReleaseTexture("Texture Room");
  _mediaPool.ReleaseTexture("Texture Door");
  _mediaPool.ReleaseTexture("Texture Room");
}

void VLGUnitScene_StartRoom::UpdateScene()
{
  _fireAnim.iSpriteAImage()->StepFrame(5);

  if(_doorOrder == 1) {
    if(_doorCnt < (_areaDoor[1].x - _areaDoor[0].x))
      _doorCnt += 4;
    else {
      _doorOrder  = 0;
      _doorOpened = true;
    }
  }
  else if(_doorOrder == -1) {
    _doorOpened = false;
    if(_doorCnt > 0) 
      _doorCnt -= 4;
    else
      _doorOrder  = 0;
  }
}

bool VLGUnitScene_StartRoom::PassableRegion(const TFloatXYZ &coord, const TFloatXYZ &radius) const
{
  const float &x = coord.x;
  const float &z = coord.z;
  const float  r = (radius.x + radius.y + radius.z) / 3.0f;

  // "room's floor" area
  if(x>=_areaFloor[0].x+r && z>=_areaFloor[0].z+r &&
     x<=_areaFloor[1].x-r && z<=_areaFloor[1].z-r)
    return(true);

  // "door" area
  if(!_doorOpened && 
     x>=_areaDoor[0].x+r && z>=_areaDoor[0].z-r &&
     x<=_areaDoor[1].x-r && z<=_areaDoor[1].z+r)
    return(false);

  // "path" area
  if(x>=_areaPath[0].x+r && z>=_areaPath[0].z   &&
     x<=_areaPath[1].x-r && z<=_areaPath[1].z+r)
    return(true);

  return(false);
}

void VLGUnitScene_StartRoom::RenderAllSolidObjects(const CMatrix3D &cameraMatrix)
{
  //_room.TNRender(cameraMatrix);
  //_room.FogTNRender(cameraMatrix);
  _room.FogTGRender(cameraMatrix);

  CMatrix3D mdoor;
  mdoor.SetToTranslationMatrix(_doorCnt, 0, 0);
  mdoor.MultMatrixCCN(cameraMatrix);
  _door.FogTGRender(mdoor);

  _path.FogTGRender(cameraMatrix);
  _soil.FogCFRender(cameraMatrix);
}

void VLGUnitScene_StartRoom::RenderAllTransparentObjects(const CMatrix3D &cameraMatrix)
{
  _fireAnim.iSprite2D().ARender(cameraMatrix);
}

bool VLGUnitScene_StartRoom::InDoorAccessArea(const TFloatXYZ &coord, const TFloatXYZ &radius) const
{
  const float &x = coord.x;
  const float &z = coord.z;
  const float  r = (radius.x + radius.y + radius.z) / 3.0f;

  // "door" area
  if(x>=_areaDoor[0].x+r && z>=_areaDoor[0].z-r &&
     x<=_areaDoor[1].x-r && z<=_areaDoor[1].z+r)
    return(false);

  // "door's access" area
  if(_doorOpened) {
    if(x>=_areaDoor[0].x-10+ 70 && z>=_areaDoor[0].z-20-r &&
       x<=_areaDoor[1].x+10+100 && z<=_areaDoor[1].z+20+r)
      return(true);
  }
  else {
    if(x>=_areaDoor[0].x-10 && z>=_areaDoor[0].z-20-r &&
       x<=_areaDoor[1].x+10 && z<=_areaDoor[1].z+20+r)
      return(true);
  }

  return(false);
}
