// ############################################################################################
//  Yuv.cc
//  A demo application for VortexGE
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// Common headers for demo application
#include "DemoCommon.h"


// Just for shortcut
typedef CBasicApplication<CEventHandler> MyApplicationBase;

// Specialized basic application class
class MyApplication : public MyApplicationBase {
  public:
    MyApplication();
    virtual void EventKeyPress(EKeySymbol eKeySym);
    virtual void EventUpdateFrame(bool bWindowIsActive);
    virtual void EventCloseWindow();

  private:
    CImageBuilder csIB1, csIB2;
    CImage        csIOrg, csIYUV, csIBck;
    CImage        csIFlA, csIFlB, csIFlC;
    CImage        csIFlD, csIFlE, csIFlF;
    CImage        csIFlG, csIFlH, csIFlI;
    CImage        csIFlJ;
    uint8_t       un8Seq;
};


// Main program
int main()
{
  MyApplication csMyApp;
  return(csMyApp.DoMainLoop());
}


// MyApplication implementation
MyApplication::MyApplication()
: MyApplicationBase("Demo : YUV Image", NULL, 640, 480, 0, 0, EAIM_Disabled),
  un8Seq(0)
{
  // Tell the user to be patient
  ShowInitSystemPleaseWaitMessage(iVideoEnv());

  // Load source image
  csIB1.LoadBMP("../media/image/back1.bmp");
  csIB1.ResizeToFitBackground();
  csIB1.BlurEffect(1);
  csIB1.TransferToImage(csIOrg);
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_rgb5-u.vib", EICM_None);
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_rgb5-c.vib", EICM_ZlibDeflate);

  //csIB1.__SaveVIB100("../.mediaCaches/yuvdemo_rgb5-u.vib");
  //csIB1.__SaveVIB100("../.mediaCaches/yuvdemo_rgb5-c.vib");
  //csIB1.__LoadVIB100("../.mediaCaches/yuvdemo_rgb5-u.vib");
  //csIB1.__LoadVIB100("../.mediaCaches/yuvdemo_rgb5-c.vib");
  //csIB1.__LoadVIB110("../.mediaCaches/yuvdemo_rgb5-u.vib");
  //csIB1.__LoadVIB110("../.mediaCaches/yuvdemo_rgb5-c.vib");

  // Convert RGB to YUV
  csIB1.ConvertRGB2YUV();
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_fyuv-u.vib", EICM_None);
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_fyuv-c.vib", EICM_ZlibDeflate);

  //csIB1.__SaveVIB110("../.mediaCaches/yuvdemo_fyuv-u.vib", EICM_None);
  //csIB1.__SaveVIB110("../.mediaCaches/yuvdemo_fyuv-c.vib", EICM_ZlibDeflate);
  //csIB1.__LoadVIB110("../.mediaCaches/yuvdemo_fyuv-u.vib");
  //csIB1.__LoadVIB110("../.mediaCaches/yuvdemo_fyuv-c.vib");
  //csIB1.__LoadVIB100("../.mediaCaches/yuvdemo_fyuv-u.vib");

  // Convert YUV to YV12
  csIB1.LoadVIB("../.mediaCaches/yuvdemo_fyuv-c.vib");
  csIB1.ConvertYUV2YV12();
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_yv12-u.vib", EICM_None);
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_yv12-c.vib", EICM_ZlibDeflate);

  // Convert YV12 to YUV
  csIB1.LoadVIB("../.mediaCaches/yuvdemo_yv12-c.vib");
  csIB1.ConvertYV122YUV();
  csIB1.TransferToImage(csIYUV);

  // Make a copy of the YUV image
  csIB2.CreateEmptyImage(csIB1);
  csIB1.CopyImage(csIB2);

  // Convert YUV to RGB
  csIB1.ConvertYUV2RGB();
  csIB1.TransferToImage(csIBck);
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_rest-u.vib", EICM_None);
  csIB1.SaveVIB("../.mediaCaches/yuvdemo_rest-c.vib", EICM_ZlibDeflate);

  // Display the file sizes
  VInfo(_NFL_, "Size File");
  system("ls -hs1 ../.mediaCaches/yuvdemo*.vib");

  // Perform some YUV specific filters
  csIB2.CopyImage(csIB1); csIB1.YUVShiftY(5);     csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlA);
  csIB2.CopyImage(csIB1); csIB1.YUVShiftY(-5);    csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlB);
  csIB2.CopyImage(csIB1); csIB1.YUVShiftU(5);     csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlC);
  csIB2.CopyImage(csIB1); csIB1.YUVShiftU(-5);    csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlD);
  csIB2.CopyImage(csIB1); csIB1.YUVShiftV(5);     csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlE);
  csIB2.CopyImage(csIB1); csIB1.YUVShiftV(-5);    csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlF);
  csIB2.CopyImage(csIB1); csIB1.YUVScaleY(2.0f);  csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlG);
  csIB2.CopyImage(csIB1); csIB1.YUVScaleY(0.5f);  csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlH);
  csIB2.CopyImage(csIB1); csIB1.YUVScaleUV(5.0f); csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlI);
  csIB2.CopyImage(csIB1); csIB1.YUVScaleUV(0.2f); csIB1.ConvertYUV2RGB(); csIB1.TransferToImage(csIFlJ);

  // Free images
  csIB1.FreeImage();
  csIB2.FreeImage();
}

void MyApplication::EventKeyPress(EKeySymbol eKeySym)
{
  switch(eKeySym) {
    case VKS_ESC : EventCloseWindow(); break;
    case ' '     : un8Seq = (un8Seq < 12) ? (un8Seq + 1) : 0; break;
    default      : break;
  }
}

void MyApplication::EventUpdateFrame(bool bWindowIsActive)
{
  // Update frame (if the application is not active)
  if(!bWindowIsActive) {
    iVideoEnv().PutRenderImageOnDisplay();
    iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
    iVideoEnv().UseDefaultBigFont();
    iVideoEnv().DrawText(0, 150, "Inactive State", ETA_CenterScreen);
    iVideoEnv().UseDefaultSmallFont();
    iVideoEnv().UpdateDisplay();
    return;
  }

  // Display the background image
  switch(un8Seq) {
    case  0 : i3DEnv().ClearRenderBuffer(csIOrg); break;
    case  1 : i3DEnv().ClearRenderBuffer(csIYUV); break;
    case  2 : i3DEnv().ClearRenderBuffer(csIBck); break;
    case  3 : i3DEnv().ClearRenderBuffer(csIFlA); break;
    case  4 : i3DEnv().ClearRenderBuffer(csIFlB); break;
    case  5 : i3DEnv().ClearRenderBuffer(csIFlC); break;
    case  6 : i3DEnv().ClearRenderBuffer(csIFlD); break;
    case  7 : i3DEnv().ClearRenderBuffer(csIFlE); break;
    case  8 : i3DEnv().ClearRenderBuffer(csIFlF); break;
    case  9 : i3DEnv().ClearRenderBuffer(csIFlG); break;
    case 10 : i3DEnv().ClearRenderBuffer(csIFlH); break;
    case 11 : i3DEnv().ClearRenderBuffer(csIFlI); break;
    case 12 : i3DEnv().ClearRenderBuffer(csIFlJ); break;
    default : break;
  }

  // Update display
  iVideoEnv().PutRenderImageOnDisplay();
  iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
  iVideoEnv().DrawText(10, 20, "Press SPACE to display the next image");
  switch(un8Seq) {
    case  0 : iVideoEnv().DrawText(10, 468, "[Original RGB Image]");            break;
    case  1 : iVideoEnv().DrawText(10, 468, "[RGB -> YUV Result]");             break;
    case  2 : iVideoEnv().DrawText(10, 468, "[YUV -> RGB Result]");             break;
    case  3 : iVideoEnv().DrawText(10, 468, "[YUV -> Y(+5) -> RGB Result]");    break;
    case  4 : iVideoEnv().DrawText(10, 468, "[YUV -> Y(-5) -> RGB Result]");    break;
    case  5 : iVideoEnv().DrawText(10, 468, "[YUV -> U(+5) -> RGB Result]");    break;
    case  6 : iVideoEnv().DrawText(10, 468, "[YUV -> U(-5) -> RGB Result]");    break;
    case  7 : iVideoEnv().DrawText(10, 468, "[YUV -> V(+5) -> RGB Result]");    break;
    case  8 : iVideoEnv().DrawText(10, 468, "[YUV -> V(-5) -> RGB Result]");    break;
    case  9 : iVideoEnv().DrawText(10, 468, "[YUV -> Y(x2.0) -> RGB Result]");  break;
    case 10 : iVideoEnv().DrawText(10, 468, "[YUV -> Y(x0.5) -> RGB Result]");  break;
    case 11 : iVideoEnv().DrawText(10, 468, "[YUV -> UV(x5.0) -> RGB Result]"); break;
    case 12 : iVideoEnv().DrawText(10, 468, "[YUV -> UV(x0.2) -> RGB Result]"); break;
    default : break;
  }
  iVideoEnv().DrawText(10, 50, "Press escape to exit ...");
  iVideoEnv().UpdateDisplay();
}

void MyApplication::EventCloseWindow()
{
  iVideoEnv().MessageBox("End Program", "Click \"OK\" to exit ...");
  MyApplicationBase::EventCloseWindow();
}
