// ############################################################################################
//  tut01a.cc
//  A tutorial application for VortexGE
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// The VortexGE include file
#include <VMain.h>


// Objects
static CVideoEnv csVideoEnv;
static C3DEnv    cs3DEnv;
static CAudioEnv &csAudioEnv = CAudioEnv::AutoDetect();


// Main program
int main()
{
  // Local variables
  uint64_t un64Frames, un64Ticks1, un64Ticks2;
  float    fFPS = 0, fOldFPS = 1;
  char     acBuff[PATH_MAX];

  // Initialize video
  csVideoEnv.CreateMainWindow(640, 480, "Tutorial 1a : Initializing the System");
  csVideoEnv.CreateRenderImage(640, 480);
  csVideoEnv.CreateRenderBuffer(640, 480);
  cs3DEnv.Initialize(&csVideoEnv);
  csVideoEnv.SetForegroundColor(csVideoEnv.CreateColor(255, 192, 255));

  // Initialize audio
  csAudioEnv.QueryDeviceName(acBuff, sizeof(acBuff));
  csAudioEnv.OpenDevice(acBuff);
  csAudioEnv.InitializeDevice();
  csAudioEnv.ActivateDevice();

  // Main loop
  csVideoEnv.AutoRepeatOff();
  un64Frames = 0;
  un64Ticks1 = cs3DEnv.GetTicks();
  while(1) {
    // Calculate FPS
    fFPS = cs3DEnv.CalcFPS();
    if(fOldFPS != fFPS) {
      fOldFPS = fFPS;
      sprintf(acBuff, "%05.1f fps", fFPS);
    }
    un64Frames++;

    // Update everything if the application is active
    if(csVideoEnv.IsActive()) {
      cs3DEnv.ClearRenderBuffer(V_RGB(12, 12, 16));
      cs3DEnv.ClearZBuffer();
      csVideoEnv.PutRenderImageOnDisplay();
      csVideoEnv.DrawText(10, 20, acBuff);
      csVideoEnv.DrawText(10, 50, "Click the left mouse button to exit ...");
      csVideoEnv.UpdateDisplay();

      // Update sound
      csAudioEnv.UpdateBuffer();

      // Pump event
      csVideoEnv.DispatchEvent();
      if(csVideoEnv.CheckEvent()) {
        // Keyboard event
        if(csVideoEnv.EventKeyPress())
          VPrint("KeyPress   : %05u (%05lu)\n", csVideoEnv.GetKeySym(),
                                                csVideoEnv.GetRawKeyCode());
        else if(csVideoEnv.EventKeyRelease())
          VPrint("KeyRelease : %05u (%05lu)\n", csVideoEnv.GetKeySym(),
                                                csVideoEnv.GetRawKeyCode());
        // Mouse event
        else if(csVideoEnv.EventButtonPress())
          VPrint("BtnPress   : %d\n", csVideoEnv.GetButtonSym());
        else if(csVideoEnv.EventButtonRelease()) {
          VPrint("BtnRelease : %d\n", csVideoEnv.GetButtonSym());
          if(csVideoEnv.GetButtonSym() == VBS_MB1) break;
        }
        // Other event
        else if(csVideoEnv.EventCloseWindow())
          break;
      }
    }
    else {
      while(csVideoEnv.IsInactive()) {
        // Update display (only of the application window is visible)
        if(csVideoEnv.IsVisible()) {
          csVideoEnv.PutRenderImageOnDisplay();
          csVideoEnv.UseDefaultBigFont();
          csVideoEnv.DrawText(0, 150, "Inactive State", ETA_CenterScreen);
          csVideoEnv.UseDefaultSmallFont();
          csVideoEnv.UpdateDisplay();
        }

        // Pump event
        csVideoEnv.DispatchEvent();
        usleep(100000);
      }
    }
  }
  un64Ticks2 = cs3DEnv.GetTicks();

  // Deactivate audio
  csAudioEnv.DeactivateDevice();

  // Calculate & display the avegare frame rate
  sprintf(acBuff, "Average frame rate = %.2f fps\n\nClick \"OK\" to exit ...",
          cs3DEnv.CalcAvgFPS(un64Frames, un64Ticks1, un64Ticks2));
  csVideoEnv.MessageBox("End Program", acBuff);
  VInfo(_N_, _F_, _L_, "Program ended, average frame rate = %.2f FPS",
        cs3DEnv.CalcAvgFPS(un64Frames, un64Ticks1, un64Ticks2));

  // OK !
  return(0);
}
