/** \file VegetationFractals.h */
// ############################################################################################
//  VegetationFractals.h
//  Definition & implementation of various CIndefiniteAreaFractal classes
//  specialized to imitate plants and leaves
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Fractal Vision : Put Fractals to Work for You
//    Dick Oliver
//    Sams Publishing
//    1992
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___FRACTAL___VEGETATION_FRACTALS_H_
#define _2D_MANIPULATION___FRACTAL___VEGETATION_FRACTALS_H_

// VortexGE headers
#include "IndefiniteAreaFractals.h"


namespace vge {

  //!
  //! \brief Fractal that will try to imitate bamboo
  //!
  //! \api experiment
  class VGE_EXPORT FractalBamboo : public CSeedReplicatorFractal {
    public:
      FractalBamboo();

    protected:
      // Parameter-control functions
      virtual bool    CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
      virtual bool    ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
      virtual uint8_t ChaosFactor() const { return(3); }
  };


  //!
  //! \brief Fractal that will try to imitate fern
  //!
  //! \api experiment
  class VGE_EXPORT FractalFern : public CSeedReplicatorFractal {
    public:
      FractalFern();

    protected:
      // Parameter-control functions
      virtual uint32_t TotalReplicationLevel() const { return(8); }
      virtual bool     CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
      virtual bool     ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
  };


  //!
  //! \brief Fractal that will try to imitate a maple leaf
  //!
  //! \api experiment
  class VGE_EXPORT FractalMapleLeaf : public CSeedReplicatorFractal {
    public:
      FractalMapleLeaf();

    protected:
      // Parameter-control functions
      virtual uint32_t TotalReplicationLevel()     const { return(6); }
      virtual uint32_t CosmosStartLevelThreshold() const { return(2); }
      virtual bool     CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
      virtual bool     ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart);
  };

} // namespace vge
#endif
