/** \file Color.h */
// ############################################################################################
//  Color.h
//  Prototype & implementation of named color functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___COLOR_H_
#define _VIDEO_ENVIRONMENT___COLOR_H_

// VortexGE header
#include "../Base/VBasics.h"

namespace vge {

  // Colors
  typedef uint32_t XColor15; //!< Color type used by the 3D environment
  typedef uint32_t XColor24; //!< Color type used by the video environment

  // Color constants
  const XColor24 INVALID_COLOR24_VALUE = uint32_t(0xFFFFFFFF);  //!< Constant for indicating invalid color value
  const XColor15 INVALID_COLOR15_VALUE = uint32_t(0xFFFFFFFF);  //!< Constant for indicating invalid color value

  const uint8_t  MAX_COLOR_VALUE            = uint8_t (31);         //!< Maximum value for each element in RGB555 color
  const uint32_t EXCLUDED_COLOR_FIRST_PIXEL = uint32_t(0xFFFFFFFF); //!< Constants to indicate that the first pixel's color will be used as color key.


  //!
  //! \brief A class containing named color constants
  //!
  //! \api pretty stable
  class VGE_EXPORT VColor {
    public:
      static const XColor15 AliceBlue;
      static const XColor15 AntiqueWhite;
      static const XColor15 Aqua;
      static const XColor15 Aquamarine;
      static const XColor15 Azure;
      static const XColor15 Beige;
      static const XColor15 Bisque;
      static const XColor15 Black;
      static const XColor15 BlanchedAlmond;
      static const XColor15 Blue;
      static const XColor15 BlueViolet;
      static const XColor15 Brown;
      static const XColor15 BurlyWood;
      static const XColor15 CadetBlue;
      static const XColor15 Chartreuse;
      static const XColor15 Chocolate;
      static const XColor15 Coral;
      static const XColor15 CornFlowerBlue;
      static const XColor15 CornSilk;
      static const XColor15 Crimson;
      static const XColor15 Cyan;
      static const XColor15 DarkBlue;
      static const XColor15 DarkCyan;
      static const XColor15 DarkGoldenrod;
      static const XColor15 DarkGray;
      static const XColor15 DarkGreen;
      static const XColor15 DarkKhaki;
      static const XColor15 DarkMagenta;
      static const XColor15 DarkOliveGreen;
      static const XColor15 DarkOrange;
      static const XColor15 DarkOrchid;
      static const XColor15 DarkRed;
      static const XColor15 DarkSalmon;
      static const XColor15 DarkSeaGreen;
      static const XColor15 DarkSlateBlue;
      static const XColor15 DarkSlateGray;
      static const XColor15 DarkTurquoise;
      static const XColor15 DarkViolet;
      static const XColor15 DeepPink;
      static const XColor15 DeepSkyBlue;
      static const XColor15 DimGray;
      static const XColor15 DodgerBlue;
      static const XColor15 Firebrick;
      static const XColor15 FloralWhite;
      static const XColor15 ForestGreen;
      static const XColor15 Fuchsia;
      static const XColor15 Gainsboro;
      static const XColor15 GhostWhite;
      static const XColor15 Gold;
      static const XColor15 Goldenrod;
      static const XColor15 Gray;
      static const XColor15 Green;
      static const XColor15 GreenYellow;
      static const XColor15 Honeydew;
      static const XColor15 HotPink;
      static const XColor15 IndianRed;
      static const XColor15 Indigo;
      static const XColor15 Ivory;
      static const XColor15 Khaki;
      static const XColor15 Lavender;
      static const XColor15 LavenderBlush;
      static const XColor15 LawnGreen;
      static const XColor15 LemonChiffon;
      static const XColor15 LightBlue;
      static const XColor15 LightCoral;
      static const XColor15 LightCyan;
      static const XColor15 LightGoldenrodYellow;
      static const XColor15 LightGreen;
      static const XColor15 LightGray;
      static const XColor15 LightPink;
      static const XColor15 LightSalmon;
      static const XColor15 LightSeaGreen;
      static const XColor15 LightSkyBlue;
      static const XColor15 LightSlateGray;
      static const XColor15 LightSteelBlue;
      static const XColor15 LightYellow;
      static const XColor15 Lime;
      static const XColor15 LimeGreen;
      static const XColor15 Linen;
      static const XColor15 Magenta;
      static const XColor15 Maroon;
      static const XColor15 MediumAquamarine;
      static const XColor15 MediumBlue;
      static const XColor15 MediumOrchid;
      static const XColor15 MediumPurple;
      static const XColor15 MediumSeaGreen;
      static const XColor15 MediumSlateBlue;
      static const XColor15 MediumSpringGreen;
      static const XColor15 MediumTurquoise;
      static const XColor15 MediumVioletRed;
      static const XColor15 MidnightBlue;
      static const XColor15 MintCream;
      static const XColor15 MistyRose;
      static const XColor15 Moccasin;
      static const XColor15 NavajoWhite;
      static const XColor15 Navy;
      static const XColor15 OldLace;
      static const XColor15 Olive;
      static const XColor15 OliveDrab;
      static const XColor15 Orange;
      static const XColor15 OrangeRed;
      static const XColor15 Orchid;
      static const XColor15 PaleGoldenrod;
      static const XColor15 PaleGreen;
      static const XColor15 PaleTurquoise;
      static const XColor15 PaleVioletRed;
      static const XColor15 PapayaWhip;
      static const XColor15 PeachPuff;
      static const XColor15 Peru;
      static const XColor15 Pink;
      static const XColor15 Plum;
      static const XColor15 PowderBlue;
      static const XColor15 Purple;
      static const XColor15 Red;
      static const XColor15 RosyBrown;
      static const XColor15 RoyalBlue;
      static const XColor15 SaddleBrown;
      static const XColor15 Salmon;
      static const XColor15 SandyBrown;
      static const XColor15 SeaGreen;
      static const XColor15 Seashell;
      static const XColor15 Sienna;
      static const XColor15 Silver;
      static const XColor15 SkyBlue;
      static const XColor15 SlateBlue;
      static const XColor15 SlateGray;
      static const XColor15 Snow;
      static const XColor15 SpringGreen;
      static const XColor15 SteelBlue;
      static const XColor15 Tan;
      static const XColor15 Teal;
      static const XColor15 Thistle;
      static const XColor15 Tomato;
      static const XColor15 Turquoise;
      static const XColor15 Violet;
      static const XColor15 Wheat;
      static const XColor15 White;
      static const XColor15 WhiteSmoke;
      static const XColor15 Yellow;
      static const XColor15 YellowGreen;
  };


  // Color converters for VortexGE render buffer
  //!
  //! Use this to create the correct XColor15 color code from the given R, G and B components.
  //!
  //! \api pretty stable
  inline XColor15 V_RGB(uint8_t un8R, uint8_t un8G, uint8_t un8B)
  { return((uint32_t(un8R) << 19) | (uint32_t(un8G) << 11) | (uint32_t(un8B) << 3)); }
  //!
  //! Use this to extract the color components from an XColor15 color code.
  //!
  //! \api pretty stable
  inline void V_RGB_Extract(uint8_t &un8R, uint8_t &un8G, uint8_t &un8B, XColor15 xColorVRGB)
  {
    un8R = ((xColorVRGB & 0x00F80000) >> 19);
    un8G = ((xColorVRGB & 0x0000F800) >> 11);
    un8B = ((xColorVRGB & 0x000000F8) >>  3);
  }

  // RGB <-> YUV converters
  //!
  //! Use this to create convert from RGB -> YUV (using the ITU-R BT.601 standard).
  //!
  //! \api pretty stable
  void VGE_EXPORT BT601_RGB2YUV(uint8_t un8R, uint8_t un8G, uint8_t un8B, uint8_t &un8Y, uint8_t &un8U, uint8_t &un8V);
  //!
  //! Use this to create convert from YUV -> RGB (using the ITU-R BT.601 standard).
  //!
  //! \api pretty stable
  void VGE_EXPORT BT601_YUV2RGB(uint8_t un8Y, uint8_t un8U, uint8_t un8V, uint8_t &un8R, uint8_t &un8G, uint8_t &un8B);
  //!
  //! Use this to create convert from RGB -> YUV (using VortexGE 15-bit color format).
  //!
  //! \api pretty stable
  void VGE_EXPORT V_RGB2YUV(uint8_t un8R, uint8_t un8G, uint8_t un8B, uint8_t &un8Y, uint8_t &un8U, uint8_t &un8V);
  //!
  //! Use this to create convert from YUV -> RGB (using VortexGE 15-bit color format).
  //!
  //! \api pretty stable
  void VGE_EXPORT V_YUV2RGB(uint8_t un8Y, uint8_t un8U, uint8_t un8V, uint8_t &un8R, uint8_t &un8G, uint8_t &un8B);

} // namespace vge
#endif
