// ############################################################################################
//  bench.h
//  Benchmarking application standard header file
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _RAW_BENCHMARK___BENCH_H_
#define _RAW_BENCHMARK___BENCH_H_

#ifndef _XOPEN_SOURCE
#  define _XOPEN_SOURCE 600
#endif

// Include the VortexGE configuration header file
#include <VConfig.h>

// C++ enabled C headers
#include <cerrno>
#include <cmath>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>
#include <cstring>
using namespace std;

// Unix standard headers
#include <stdint.h>
#include <unistd.h>

// Unix system headers
#include <sys/time.h>


//
// Macros
//

// Optimization macros
#if defined(__GNUC__) && defined(__USE_ASM__)
#  if defined(__i386__)
#    define _ASM_A32_
#  endif
#  if defined(__x86_64__)
#    define _ASM_A64_
#  endif
#endif

#if defined(_ASM_A32_) || defined(_ASM_A64_)
#  define _ASM_ASM_
#  if defined(__USE_MMX__)
#    define _ASM_MMX_
#  endif
#  if defined(__USE_SSE__)
#    define _ASM_SSE_
#    define _ALIGN_128_BITS_
#  endif
#  if defined(__USE_SSE2__)
#    define _ASM_SSE2_
#    define _ALIGN_128_BITS_
#  endif
#  if defined(__USE_SSE3__)
#    define _ASM_SSE3_
#    define _ALIGN_128_BITS_
#  endif
#  if defined(__USE_3DNOW__)
#    define _ASM_3DNOW_
#  endif
#endif

// Alignment directive for buffer allocation
#if defined(__USE_ALIGNED_MEMOPS__) && defined(_ALIGN_128_BITS_)
#  define _ALIGN_BUFF_MEM_
#endif

// Directive to make the variable's memory address is
// specially aligned for SSE support
#if defined(_ASM_SSE_) || defined(_ASM_SSE2_) || defined(_ASM_SSE3_)
#  define _sse __attribute__((aligned(16)))
#else
#  define _sse
#endif

#endif
