// tbf.cc
// Tab Beautifier
//
// Copyright (C) 2004-2006 Aloysius Indrayanto
// This small utility is licensed under the GNU LGPL
//
// BUG: Will not work with multiline string
//
// Usage: ./tbf in_file  [> out-file]
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


// Make sure that this is set with the correct value
#define TAB_SIZE 2


inline bool checkBreakedLine(char c)
{
  return(c==',' ||
         c=='|' || c=='&' || c=='^' ||
         c=='+' || c=='-' || c=='*' || c=='/' || c=='%');
}


int main(int argc, char *argv[])
{
  if(argc != 2) {
    printf("\nUsage:");
    printf("\n  %s in-file [> out-file]", argv[0]);
    printf("\n\n");
    exit(1);
  }

  FILE *fp = fopen(argv[1], "rt");
  if(!fp) {
    printf("\nCouldn't open input file \"%s\"\n\n", argv[1]);
    exit(1);
  }

  bool inString    = false;
  bool breakedLine = false;
  int  tabCount    = 0;

  char   *line;
  size_t r;
  int    n;
  for(;;) {
    line = NULL;
    if((n = getline(&line, &r, fp)) == -1) break;

    // Remove line that only contains spaces and/or tabs
    bool empty = true;
    for(int i = 0; i < n; i++) {
      char c = line[i];
      if(c!=0 && c!='\n' && c!='\r' && c!='\t' && c!=' ') {
        empty = false;
        break;
      }
    }
    if(empty) {
      fputc('\n', stdout);
      continue;
    }

    // Remove trailing spaces/tabs
    int z = n;
    for(int i = (n-1); i >= 0; i--) {
      char c = line[i];
      if(c!=0 && c!='\n' && c!='\r' && c!='\t' && c!=' ') break;
      line[i] = 0;
      z--;
    }

    // Process the line
    bool gotNL = false;
    if(breakedLine) {
      int tc = 0;
      for(int i = 0; i < z; i++) {
        char c = line[i];
        // If it is a tab character only print if the total number of tabs
        // is still smaller than the total number of tabs of the previous
        // line, otherwise just print some spaces
        if(!inString && c=='\t') {
          tc++;
          if(tc <= tabCount)
            fputc('\t', stdout);
          else {
            for(int s = 0; s < TAB_SIZE; s++) fputc(' ', stdout);
          }
        }
        // Just prints other characters
        else {
          for(int d = i; d < z; d++) {
            char c = line[d];
            //if(c=='\'' || c=='\"') {
            //  if(d==0 || line[d-1] != '\\') inString = !inString;
            //}
            fputc(c, stdout);
          }
          break;
          }
      }
      if(!inString) breakedLine = checkBreakedLine(line[z-1]);
    }
    else {
      // Count the number of preceding tabs and print the processed characters
      tabCount = 0;
      for(int i = 0; i < z; i++) {
        char c = line[i];
        if(!inString && c=='\t') {
          fputc('\t', stdout);
          tabCount++;
        }
        else {
          for(int d = i; d < z; d++) {
            char c = line[d];
            //if(c=='\'' || c=='\"') {
            //  if(d==0 || line[d-1] != '\\') inString = !inString;
            //}
            fputc(c, stdout);
          }
          break;
        }
      }
      if(!inString) breakedLine = checkBreakedLine(line[z-1]);
    }
    fputc('\n', stdout);

    free(line);
  }

  fclose(fp);
}
